def build_kerning_classes(source):
    """Build kerning classes from source font file."""
    kerning_classes = {}
    for glyph in source.glyphs:

        if glyph.leftKerningGroup:
            first_group = f"@MMK_R_{glyph.leftKerningGroup}"
            if first_group not in kerning_classes:
                kerning_classes[first_group] = []
            if glyph.name not in kerning_classes[first_group]:
                kerning_classes[first_group].append(glyph.name)

        if glyph.rightKerningGroup:
            second_group = f"@MMK_L_{glyph.rightKerningGroup}"
            if second_group not in kerning_classes:
                kerning_classes[second_group] = []
            if glyph.name not in kerning_classes[second_group]:
                kerning_classes[second_group].append(glyph.name)

    return kerning_classes


def unicodes_for_glyph_or_class(font, glyph_or_class, kerning_classes):
    """Return first found unicode for glyph or class."""
    if glyph_or_class.startswith("@"):
        unicodes = []
        for glyph_name in kerning_classes[glyph_or_class]:
            unicodes.extend(font.glyphs[glyph_name].unicodes)
        return unicodes
    else:
        if font.glyphs[glyph_or_class].unicodes:
            return font.glyphs[glyph_or_class].unicodes
        else:
            return []


def glyph_for_unicode(font, unicode):
    """Return glyph for unicode."""
    for glyph in font.glyphs:
        if unicode in glyph.unicodes:
            return glyph



from setuptools import setup, find_packages
import platform
import os

this_directory = os.path.dirname(__file__)

WIN = platform.system() == "Windows"
MAC = platform.system() == "Darwin"
LINUX = platform.system() == "Linux"

install_requires = [
    line.strip()
    for line in open(os.path.join(this_directory, "requirements.txt")).readlines()
    if bool(line.strip()) and not line.strip().startswith("#")
]
long_description = open(os.path.join(this_directory, "README.md")).read()
print(install_requires)

setup(
    name="kerningvalidator",  # How you named your package folder (MyLib)
    version="0.1.1.post2",  # .post2
    license="apache-2.0",
    # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    author="Yanone",  # Type in your name
    author_email="post@yanone.de",  # Type in your E-Mail
    url="https://github.com/yanone/kerningvalidator",
    # Provide either the link to your github or to your website
    keywords=["fonts"],  # Keywords that define your package best
    install_requires=install_requires,
    entry_points={
        "console_scripts": [
            "kerningvalidator = kerningvalidator.cli:main",
        ]
    },
    package_dir={"": "Lib"},
    packages=find_packages("Lib"),
    include_package_data=True,
    long_description=long_description,
    long_description_content_type="text/markdown",
)



import argparse
import kerningvalidator
import sys


def main():
    parser = argparse.ArgumentParser(description="Compare font source file kerning with font binary file kerning.")
    parser.add_argument(dest="glyphs_source", help="Glyphs source font file (.glyphs)")
    parser.add_argument(dest="binary", help="Binary font file (.ttf/.otf)")

    args = parser.parse_args()

    missing_kerning = kerningvalidator.missing_kerning(args.glyphs_source, args.binary)
    if missing_kerning:
        sys.exit(1)


if __name__ == "__main__":
    main()



import logging
import glyphsLib
from .source import build_kerning_classes, unicodes_for_glyph_or_class, glyph_for_unicode
import vharfbuzz as vhb


def buf_x_translate(hb_font, buf):
    x_translate = 0
    x_cursor = 0
    y_cursor = 0
    paths = []
    for info, pos in zip(buf.glyph_infos, buf.glyph_positions):
        dx, dy = pos.position[0], pos.position[1]
        x_translate = max(x_translate, x_cursor + dx)
        x_cursor += pos.position[2]
        y_cursor += pos.position[3]
    return x_translate


def width(hb_font, variations, string, kerning):
    """Get kerning of string."""

    hb_buf = hb_font.shape(string, {"features": {"kern": kerning}, "variations": variations})
    x_translate = buf_x_translate(hb_font, hb_buf)
    return x_translate


def shaped_kerning(hb_font, variations, string):
    """Compare width of string with and without kerning."""

    return width(hb_font, variations, string, True) - width(hb_font, variations, string, False)


def missing_kerning(glyphs_source, font_binary):
    """Compare kerning between source and binary font files."""

    problems = []

    # Open source font file
    source = glyphsLib.load(glyphs_source)

    kerning_classes = build_kerning_classes(source)
    # print(kerning_classes)

    # Harfbuzz
    hb_font = vhb.Vharfbuzz(font_binary)

    # Cycle through kerning dictionaries
    for direction, kerning_data in (
        ("LTR", source.kerningLTR),
        ("RTL", source.kerningRTL),
    ):
        for master_id in kerning_data:

            # Assemble variations matrix
            variations = {}
            for i, axis in enumerate(source.axes):
                variations[axis.axisTag] = source.masters[master_id].axes[i]
            print(f"Checking master: variations: {variations}, direction: {direction}")

            for first in kerning_data[master_id]:

                # Complain about missing first glyph or class
                if first.startswith("@"):
                    if first not in kerning_classes:
                        logging.warning(f"Kerning found but class not defined for first in pair: {first}")
                else:
                    if first not in source.glyphs:
                        logging.warning(f"Kerning found but glyph not defined for first in pair: {first}")

                # Found first glyph or class
                if first in kerning_classes or first in source.glyphs:

                    for second in kerning_data[master_id][first]:

                        # Complain about missing second glyph or class
                        if second.startswith("@"):
                            if second not in kerning_classes:
                                logging.warning(f"Kerning found but class not defined for second in pair: {second}")
                        else:
                            if second not in source.glyphs:
                                logging.warning(f"Kerning found but glyph not defined for second in pair: {second}")

                        # Found second glyph or class
                        if second in kerning_classes or second in source.glyphs:
                            first_unicodes = unicodes_for_glyph_or_class(source, first, kerning_classes)
                            second_unicodes = unicodes_for_glyph_or_class(source, second, kerning_classes)

                            # Cycle through all unicode combinations
                            for first_unicode in first_unicodes:
                                first_glyph = glyph_for_unicode(source, first_unicode)
                                for second_unicode in second_unicodes:
                                    second_glyph = glyph_for_unicode(source, second_unicode)
                                    found = shaped_kerning(
                                        hb_font,
                                        variations,
                                        f"{chr(int(first_unicode, 16))}{chr(int(second_unicode, 16))}",
                                    )

                                    # Class kerning
                                    expected = kerning_data[master_id][first][second]
                                    difference = found - expected

                                    if difference < -1 or difference > 1:
                                        # Kerning exceptions
                                        if first_glyph and second_glyph:
                                            first_name = first_glyph.name
                                            second_name = second_glyph.name

                                            if (
                                                first_name in kerning_data[master_id]
                                                and second_name in kerning_data[master_id][first_name]
                                            ):
                                                expected = kerning_data[master_id][first_name][second_name]

                                            elif (
                                                first in kerning_data[master_id]
                                                and second_name in kerning_data[master_id][first]
                                            ):
                                                expected = kerning_data[master_id][first][second_name]

                                            elif (
                                                first_name in kerning_data[master_id]
                                                and second in kerning_data[master_id][first_name]
                                            ):
                                                expected = kerning_data[master_id][first_name][second]

                                    difference = found - expected

                                    # Allow 1 unit difference for rounding errors
                                    if difference < -1 or difference > 1:
                                        first_unicode_as_str = f"{int(first_unicode, 16):#0{6}x}"
                                        second_unicode_as_str = f"{int(second_unicode, 16):#0{6}x}"
                                        logging.error(
                                            f"Not in font: dir: {direction}, var: {variations}, 1st:"
                                            f" {first} ({chr(int(first_unicode, 16))} {first_unicode_as_str} '{first_glyph.name}'),"
                                            " 2nd:"
                                            f" {second} ({chr(int(second_unicode, 16))} {second_unicode_as_str} '{second_glyph.name}'),"
                                            f" expected: {expected}, found: {found})"
                                        )
                                        problems.append(
                                            (
                                                variations,
                                                first,  # group or glyph name
                                                chr(int(first_unicode, 16)),  # encoded character
                                                int(first_unicode, 16),  # 1st unicode (decimal)
                                                second,  # group or glyph name
                                                chr(int(second_unicode, 16)),  # encoded character
                                                int(second_unicode, 16),  # 2nd unicode (decimal)
                                                expected,  # expected
                                                found,  # found
                                            )
                                        )
    return problems
